import argparse
from abc import ABC, abstractmethod
import os

import astropy
from astropy.coordinates import SkyCoord
import astropy.units as u
from astropy.io import fits
import matplotlib.pyplot as plt
import numpy as np
import pyvo
from astroquery.ipac.ned import Ned
from astroquery.vizier import Vizier
from scipy.optimize import curve_fit


plt.rcParams.update(
    {
        # "text.usetex": True,
        "font.family": "DejaVu Sans",
        "font.size": 20,
    }
)


class Fitter(ABC):
    fit_parameters = None
    fit_parameters_chi = None
    freq_points = 0

    def __init__(self, freq_ref=144e6):
        self.freq_ref = freq_ref

    @abstractmethod
    def fit_func(self, *args, **kwargs):
        pass

    def fit(self, freq, flux_density, p0, sigma):
        self.freq_points = len(freq)
        popt, pcov = curve_fit(
            self.fit_func, xdata=freq, ydata=flux_density, p0=p0, sigma=sigma
        )
        self.fit_parameters = popt
        residuals = flux_density - self.fit_func(np.array(freq), *popt)
        flux_err = sigma
        chi_sqr = np.sum(((residuals) / flux_err) ** 2)
        self.stats = chi_sqr
        return popt, chi_sqr

    def plot(
        self,
        freqs,
        fluxs,
        f_err,
        point_labels,
        file_name,
        outdir=".",
    ):
        freq_smooth = np.linspace(10e6, 10e9, 1000)
        plt.figure(figsize=(8, 8))
        plt.loglog()
        plt.gca().set(xlabel="Frequency [Hz]", ylabel="Flux density [Jy]")
        plt.setp(plt.gca().spines.values(), linewidth=1.5)
        plt.gca().xaxis.set_tick_params(width=1.5, which="both")
        plt.gca().yaxis.set_tick_params(width=1.5, which="both")
        plt.errorbar(freqs, fluxs, yerr=f_err, ls="none")
        plt.scatter(freqs, fluxs, marker="s", ec="k", fc="gray", s=64)

        if point_labels:
            for f, s, l in zip(freqs, fluxs, point_labels):
                plt.annotate(l, (f, s * 1.1), fontsize=14)
        plt.plot(
            freq_smooth,
            self.fit_func(freq_smooth, *self.fit_parameters),
            color="k",
            linestyle="--",
            linewidth=1.5,
        )
        # plt.text(0.1, 0.2, "Spectral fit:\n\n${:.2f} \\left(\\frac{{\\nu}}{{{:.0f}\\ \\mathrm{{MHz}}}}\\right)^{{{:.2f} + {:.2f} \\log_{{10}}(\\nu / {:.0f}\\ \\mathrm{{MHz}})}}$".format(self.fit_parameters[0], self.freq_ref/1e6, self.fit_parameters[1], self.fit_parameters[2], self.freq_ref/1e6), transform=plt.gca().transAxes, fontsize=16)
        plt.text(
            0.1,
            0.2,
            "Fitted parameters:\nS = {:.2f}\n$\\alpha_1$ = {:.2f}\n$\\alpha_2$ = {:.2f}\n$\\chi^2$ = {:.2f}\n$\\nu_{{\\mathrm{{ref}}}}$ = {:.2f} MHz".format(
                self.fit_parameters[0],
                self.fit_parameters[1],
                self.fit_parameters[2],
                self.stats,
                self.freq_ref / 1e6,
            ),
            transform=plt.gca().transAxes,
            fontsize=16,
            va="center_baseline",
        )
        plt.xlim(10e6, 10e9)
        plt.ylim(10e-3, 10)
        if file_name:
            plt.savefig(
                os.path.join(outdir, f"{file_name}.png"), dpi=300, bbox_inches="tight"
            )
        else:
            plt.savefig(
                os.path.join(outdir, "spectrum.png"), dpi=300, bbox_inches="tight"
            )


class LogFitter(Fitter):
    def fit_func(self, freq, I0, alpha, beta):
        return I0 * (freq / self.freq_ref) ** (
            alpha + beta * np.log10(freq / self.freq_ref)
        )


def fit_from_NED(ra: float, dec: float, radius: float, outdir: str):
    obj = Ned.query_region(f"{ra}d {dec}d", radius=radius * u.arcsec, cache=False)[
        "Object Name"
    ][0]

    ned_table = Ned.get_table(obj, table="photometry", cache=False)
    ned_photometry = ned_table[np.where(ned_table["Frequency"] < 1e10)]
    freqs = ned_photometry["Frequency"]
    fluxd_ned = ned_photometry["Flux Density"]
    fluxd_Ned_err = 0.1 * fluxd_ned  # PLACEHOLDER
    fitter = LogFitter()
    fitter.fit(freqs, fluxd_ned, p0=(1.0, -0.8, 0.0), sigma=fluxd_Ned_err)
    fitter.plot(
        freqs, fluxd_ned, file_name=f"spectrum_{ra:.3f}_{dec:.3f}_NED", outdir=outdir
    )
    return fitter


def query_vizier(catalogue: str, ra: float, dec: float, radius: float):
    v = Vizier(catalog=catalogue)
    q = v.query_region(f"{ra}, {dec}", radius=radius * u.arcsec, cache=False)
    if q:
        return q[0]
    else:
        raise RuntimeError("Source not found in requested survey.")


def query_bootstrap(catalogue: str, ra: float, dec: float, radius: float):
    with fits.open(catalogue) as hdul:
        catalog = astropy.table.Table(hdul[1].data)
    cat_coords = SkyCoord(catalog["RA"], catalog["DEC"], unit="deg")
    targ_coord = SkyCoord(ra, dec, unit="deg")
    sep = targ_coord.separation(cat_coords)
    match = sep < (radius * u.arcsec)
    q = catalog[match]
    if q:
        return q[0]
    else:
        raise RuntimeError("Source not found in requested survey.")


def query_vo(vo_server: str, ra: float, dec: float, radius: float):
    query = pyvo.dal.scs.SCSQuery(vo_server, maxrec=10)
    query["RA"] = ra
    query["DEC"] = dec
    query.radius = radius
    t = query.execute()
    return t


def fit_from_trusted_surveys(ra: float, dec: float, radius: float, outdir: str):
    frequency = []
    flux_density = []
    flux_err = []
    survey_name = []
    has_survey = 0b0000000
    try:
        s_lolss = (
            query_vizier(VIZIER_LOLSS_DR1, ra, dec, radius)["Ftot"].to("Jy").value[0]
        )
        e_s_lolss = (
            query_vizier(VIZIER_LOLSS_DR1, ra, dec, radius)["E_Total_flux"]
            .to("Jy")
            .value[0]
        )
        has_survey ^= 0b1000000
        frequency.append(60e6)
        flux_density.append(s_lolss)
        if e_s_lolss == 0 or None:
            e_s_lolss = 0.1 * s_lolss
            print("LoLSS DR1 has an invalid error. Using 10% instead")
        flux_err.append(e_s_lolss)
        survey_name.append("LOLSS")
    except RuntimeError:
        print("Source not in LoLSS DR1")

    # Old VLSSr code which we think is wrong
    # try:
    #     s_vlssr = query_vizier(VIZIER_VLSSr, ra, dec, radius)["Sp"].to("Jy").value[0]
    #     has_survey ^= 0b0100000
    #     frequency.append(74e6)
    #     flux_density.append(s_vlssr)
    #     survey_name.append("VLSSr")
    # except RuntimeError:
    #     print("Source not in VLSSr")

    try:
        s_vlssr = query_bootstrap(BOOTSTRAP_VSSLr, ra, dec, radius)["Total_flux"]  # Jy
        e_s_vlssr = query_bootstrap(BOOTSTRAP_VSSLr, ra, dec, radius)[
            "E_Total_flux"
        ]  # Jy
        frequency.append(74e6)
        flux_density.append(s_vlssr)
        if e_s_vlssr == 0 or None:
            e_s_vlssr = 0.1 * s_vlssr
            print("VLSSr has an invalid error so using 10% of flux density instead.")
        flux_err.append(e_s_vlssr)
        survey_name.append("VLSSr")
    except RuntimeError:
        print("Source not in VLSSr")

    try:
        lotss = query_vo(VO_LOTSS_DR2, ra, dec, 3.0 / 3600)
        s_lotss = lotss["Total_flux"][0] * 1e-3
        e_s_lotss = lotss["E_Total_flux"][0] * 1e-3
        has_survey ^= 0b0010000
        frequency.append(144e6)
        survey_name.append("LOTSS")
        flux_density.append(s_lotss)
        if e_s_lotss == 0 or None:
            e_s_lotss = 0.1 * s_lotss
            print(
                "LoTSS DR2 has an invalid error so using 10% of flux density instead."
            )
        flux_err.append(e_s_lotss)
        HAS_LOTSS = True
    except (RuntimeError, IndexError):
        HAS_LOTSS = False
        print("Source not in LOTSS, trying TGSS")
        try:
            s_tgss = (
                query_vizier(VIZIER_TGSS, ra, dec, radius)["Stotal"].to("Jy").value[0]
            )
            e_s_tgss = (
                query_vizier(VIZIER_TGSS, ra, dec, radius)["e_Stotal"].to("Jy").value[0]
            )
            has_survey ^= 0b0010000
            frequency.append(150e6)
            flux_density.append(s_tgss)
            if e_s_tgss == 0 or None:
                e_s_tgss = 0.1 * s_tgss
                print("TGSS has an invalid error so using 10% of flux density instead.")
            flux_err.append(e_s_tgss)
            survey_name.append("TGSS")
        except RuntimeError:
            print("Source not in TGSS")
    HAS_TGSS = not HAS_LOTSS

    try:
        s_wenss = query_vizier(VIZIER_WENSS, ra, dec, radius)["Sint"].to("Jy").value[0]
        has_survey ^= 0b0001000
        # 325 MHz in main part, 352 MHz in polar part
        if dec < 72:
            frequency.append(325e6)
        else:
            frequency.append(352e6)
        flux_density.append(s_wenss)
        e_s_wenss = 0
        if e_s_wenss == 0 or None:
            e_s_wenss = 0.2 * s_wenss
            print("WENSS has an invalid error so using 20% of flux density instead.")
        flux_err.append(e_s_wenss)  # PLACEHOLDER
        survey_name.append("WENSS")
    except RuntimeError:
        print("Source not in WENSS")

    try:
        s_sumss = query_vizier(VIZIER_SUMSS, ra, dec, radius)["St"].to("Jy").value[0]
        e_s_sumss = (
            query_vizier(VIZIER_SUMSS, ra, dec, radius)["e_St"].to("Jy").value[0]
        )
        has_survey ^= 0b0000100
        frequency.append(843e6)
        flux_density.append(s_sumss)
        if e_s_sumss == 0 or None:
            e_s_sumss = 0.1 * s_sumss
            print("SUMSS has an invalid error so using 10% of flux density instead.")
        flux_err.append(e_s_sumss)
        survey_name.append("SUMSS")
    except RuntimeError:
        print("Source not in SUMSS")

    try:
        s_first = query_vizier(VIZIER_FIRST, ra, dec, radius)["Fint"].to("Jy").value[0]
        has_survey ^= 0b0000010
        frequency.append(1.4e9)
        flux_density.append(s_first)
        # Errors reported to be within ~5%
        flux_err.append(0.05 * s_first)
        survey_name.append("FIRST")
    except RuntimeError:
        print("Source not in FIRST")

    try:
        s_gb6 = query_vizier(VIZIER_GB6, ra, dec, radius)["Flux"].to("Jy").value[0]
        e_s_gb6 = query_vizier(VIZIER_GB6, ra, dec, radius)["e_Flux"].to("Jy").value[0]
        has_survey ^= 0b0000001
        frequency.append(4.85e9)
        flux_density.append(s_gb6)
        flux_err.append(s_gb6 * 0.2)
        survey_name.append("GB6")
    except RuntimeError:
        print("Source not in GB6")

    try:
        s_vlass = (
            query_vizier(VIZIER_VLASS_QL1, ra, dec, radius)["Ftot"].to("Jy").value[0]
        )
        has_survey ^= 0b0000001
        frequency.append(3e9)
        # VLASS reports a ~15% underestimate of measurements in the QL catalogues,
        # so we roughly correct that here.
        flux_density.append(s_vlass * 1.15)
        flux_err.append(0.2 * s_vlass)  # PLACEHOLDER
        survey_name.append("VLASS")
    except RuntimeError:
        print("Source not in VLASS QL1")
    print(f"Survey mask: {has_survey:07b}")

    if HAS_LOTSS:
        fitter = LogFitter(freq_ref=144e6)
    elif HAS_TGSS:
        fitter = LogFitter(freq_ref=150e6)

    if len(flux_density) > 2:
        fitter = LogFitter(freq_ref=144e6)
        fitter.fit(frequency, flux_density, p0=(1.0, -0.8, 0.0), sigma=flux_err)
        fitter.plot(
            frequency,
            flux_density,
            flux_err,
            point_labels=survey_name,
            file_name=f"spectrum_{ra:.3f}_{dec:.3f}",
            outdir=outdir,
        )
        return fitter

    else:
        raise RuntimeError("Not Enough Data to Fit")


# 55 MHz
VIZIER_LOLSS_DR1 = "J/A+A/673/A165/lolss1g"
# 74 MHz - ???
VIZIER_VLSSr = "VIII/97/catalog"
# VLSSr_FIXED
BOOTSTRAP_VSSLr = "https://www.extragalactic.info/bootstrap/VLSS.fits"
# 325 MHz
VIZIER_WENSS = "VIII/62/wenss"
# 1.4 GHz
VIZIER_FIRST = "VIII/92/first14"
# 4.85 GHz
VIZIER_GB6 = "VIII/40/gb6"
# 843 MHz
VIZIER_SUMSS = "VIII/81B/sumss212"
# 150 MHz
VIZIER_TGSS = "J/A+A/598/A78/table3"
# 3 GHz
VIZIER_VLASS_QL1 = "J/ApJS/255/30/comp"


# 144 MHz
VO_LOTSS_DR2 = "https://vo.astron.nl/lotss_dr2/q/src_cone/scs.xml"

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Fit a synchrotron spectrum for a specific source using archival radio data."
    )
    parser.add_argument(
        "--output_dir",
        dest="outdir",
        type=str,
        help="directory to save results in [default cwd]",
        default=".",
    )
    parser.add_argument(
        "--ra", type=float, help="Right ascension of the target source."
    )
    parser.add_argument("--dec", type=float, help="Declination of the target source.")
    parser.add_argument(
        "--match_radius",
        type=float,
        help="Radius in arcsec within which to allow matching with archival sources (N.B. some surveys have a hardcoded matchin radius).",
        default=12.0,
    )
    parser.add_argument(
        "--photometry_mode",
        choices=["trusted", "NED"],
        type=str,
        help="Set to 'trusted' to only use vetted radio surveys for fitting the spectrum. Set to 'NED' to use whatever photometry is available in NED between 10 MHz and 10 GHz.",
        default="trusted",
    )
    args = parser.parse_args()
    if args.photometry_mode == "trusted":
        fit_from_trusted_surveys(args.ra, args.dec, args.match_radius, args.outdir)
    elif args.photometry_mode == "NED":
        fit_from_NED(args.ra, args.dec, args.match_radius, args.outdir)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Search VLASS for a given RA and Dec
Downloads best subtile
Original code by Anna Ho
Edited by Roland Timmerman
"""


#Imports
import os
import sys
import errno
import subprocess
import glob
import argparse
import numpy as np
from astropy.io import fits
from astropy.coordinates import SkyCoord
from astropy.wcs.utils import skycoord_to_pixel
from astropy.wcs import WCS
from astropy.nddata import Cutout2D
from importlib.resources import files
try:
    from urllib.request import urlopen
except ImportError:
    from urllib2 import urlopen

#Settings to potentially tweak
summary_file_location = files("plot_field").joinpath("VLASS_dyn_summary.php")
crop = True
crop_scale = 256
consider_QA_rejected = True


def sign(x):
    """
    Returns the mathematical sign of a number
    Input: x (int/float)
    Returns: +1 or -1
    """
    
    if x<0:
        return -1
    else:
        return 1

def get_tiles():
    """
    Read tiles from tile catalog. If file missing, try wget https://archive-new.nrao.edu/vlass/VLASS_dyn_summary.php
    Input: None
    Returns: tile names (numpy str), dec_min (numpy float), dec_max (numpy float), ra_min (numpy float), ra_max (numpy float), observing epoch (numpy str), observation date (numpy str)
    """
    
    name_array = []
    dec_min_array = []
    dec_max_array = []
    ra_min_array = []
    ra_max_array = []
    epoch_array = []
    obsdate_array = []
    
    with open(summary_file_location, 'r') as f:
        lines = f.readlines()[3:]
        for line in lines:
            name = line[0:6]
            dec_min = float(line[9:16])
            dec_max = float(line[18:25])
            ra_min = float(line[28:34])
            ra_max = float(line[37:43])
            epoch = line[47:55]
            obsdate = line[61:71]
            status = line[86:97]
            
            if status == f"100% imaged":
                name_array.append(name)
                dec_min_array.append(dec_min)
                dec_max_array.append(dec_max)
                ra_min_array.append(ra_min)
                ra_max_array.append(ra_max)
                epoch_array.append(epoch)
                obsdate_array.append(obsdate)

    name_array = np.array(name_array)
    ra_min_array = np.array(ra_min_array)
    ra_max_array = np.array(ra_max_array)
    dec_min_array = np.array(dec_min_array)
    dec_max_array = np.array(dec_max_array)
    epoch_array = np.array(epoch_array)
    obsdate_array = np.array(obsdate_array)

    return (name_array, dec_min_array, dec_max_array, ra_min_array, ra_max_array, epoch_array, obsdate_array)


def search_tiles(tiles, c):
    """
    Search the tile catalog for tiles containing the input coordinate
    Input: tile properties (see get_tiles() output), c (SkyCoord object)
    Returns: tile name (str), observing epoch (str), observing date (str)
    """
    ra_h = c.ra.hour
    dec_d = c.dec.deg
    names, dec_min, dec_max, ra_min, ra_max, epochs, obsdate = tiles
    has_dec = np.logical_and(dec_d >= dec_min, dec_d < dec_max)
    has_ra = np.logical_and(ra_h >= ra_min, ra_h < ra_max)
    in_tile = np.logical_and(has_ra, has_dec)
    name = names[in_tile]
    epoch = epochs[in_tile]
    date = obsdate[in_tile]    
    if len(name) == 0:
        raise IndexError("Zero VLASS tiles available for the given coordinate")
    c_grid = SkyCoord(7.5*(ra_min[in_tile]+ra_max[in_tile]), 0.5*(dec_min[in_tile]+dec_max[in_tile]), unit='deg', frame='icrs')
    dist = c_grid.separation(c)
    nearest_tiles = np.where(dist == np.min(dist))[0]
    best_idx = nearest_tiles[epoch==max(epoch[nearest_tiles])][0]
    return name[best_idx], epoch[best_idx], date[best_idx]


def get_subtiles(tilename, epoch, consider_QA_rejected):
    """
    For a given tile name, get the subtile filenames in the VLASS directory
    Parse those filenames and return a list of subtile RA and Dec.
    RA and Dec returned as a SkyCoord object
    """
    
    #Obtain the HTML for the given tile
    if epoch[5] == "1":
        urlpath = urlopen("https://vlass-dl.nrao.edu/vlass/quicklook/{}v2/{}".format(epoch, tilename))
    else:
        urlpath = urlopen("https://vlass-dl.nrao.edu/vlass/quicklook/{}/{}".format(epoch, tilename))
    string = urlpath.read().decode('utf-8').split("\n")

    if consider_QA_rejected:
        #Obtain the HTML for the QA Rejected
        if epoch[5] == "1":
            urlpath_rejected = urlopen("https://vlass-dl.nrao.edu/vlass/quicklook/{}v2/QA_REJECTED".format(epoch))
        else:
            urlpath_rejected = urlopen("https://vlass-dl.nrao.edu/vlass/quicklook/{}/QA_REJECTED".format(epoch))
        string += urlpath_rejected.read().decode('utf-8').split("\n")

    #Select only the subtile parts
    vals = np.array([val.strip() for val in string if ("href" in val.strip()) and (tilename in val.strip())])
    
    #Select the coordinate part. You want the 'VLASS1.1.ql.T25t12.J150000+603000.10.2048.v1/' bit
    fname = np.array([val.split("\"")[7] for val in vals])

    #Split out the actual coordinate string
    pos_raw = np.array([val.split(".")[4] for val in fname])
    if '-' in pos_raw[0]:
        # dec < 0
        ra_raw = np.array([val.split("-")[0] for val in pos_raw])
        dec_raw = np.array([val.split("-")[1] for val in pos_raw])
    else:
        # dec > 0
        ra_raw = np.array([val.split("+")[0] for val in pos_raw])
        dec_raw = np.array([val.split("+")[1] for val in pos_raw])
    ra = []
    dec = []
    for ii,val in enumerate(ra_raw):
        if val[1:3] == '24':
            rah = '00'
        else:
            rah = val[1:3]
        ra.append("{}h{}m{}s".format(rah, val[3:5], val[5:]))
        dec.append("{}d{}m{}s".format(dec_raw[ii][:2], dec_raw[ii][2:4], dec_raw[ii][4:]))
    ra = np.array(ra)
    dec = np.array(dec)
    c = SkyCoord(ra, dec, frame='icrs')#.directional_offset_by(45*u.deg, 0.75*u.deg)
    return fname, c


def get_cutout(imname, c, crop_scale):
    #Define output name
    output_fits = "vlass_poststamp.fits"
    
    #Get header info
    hdu_list = fits.open(imname)
    header = hdu_list[0].header
    data = hdu_list[0].data[0,0,:,:]
    
    #Obtain header and drop useless axes
    wcs = WCS(header)
    wcs = wcs.dropaxis(2).dropaxis(2)
    
    pixel_coords = skycoord_to_pixel(SkyCoord(c.ra.deg, c.dec.deg, unit='deg', frame='icrs'), wcs)
    
    if pixel_coords[0] < 0  or pixel_coords[1] < 0 or pixel_coords[0] > data.shape[0]  or pixel_coords[1] > data.shape[1]:
        subprocess.call("rm -f {}".format(imname), shell=True)
        raise Exception("Requested coordinate not within the available subtiles. Consider running with consider_QA_rejected=True to also search additional subtiles which failed initial QA checks")
        
    #Produce a cutout
    cutout = Cutout2D(data, c, (crop_scale, crop_scale), wcs=wcs)
        
    #Update the HDU
    hdu_list[0].data = cutout.data
    new_header = cutout.wcs.to_header()
    hdu_list[0].header.update(new_header)
    hdu_list[0].header.set('NAXIS', 4)
    hdu_list[0].header.insert('NAXIS2', ('NAXIS3', 1), after=True)
    hdu_list[0].header.insert('NAXIS3', ('NAXIS4', 1), after=True)
    hdu_list[0].header.remove('WCSAXES', ignore_missing=True)
    hdu_list[0].header.remove('MJDREF', ignore_missing=True)
    hdu_list[0].header.remove('MJD-OBS', ignore_missing=True)
    
    #Write the new fits
    hdu_list.writeto(output_fits, overwrite=True)
    
    #Cleanup
    #subprocess.call("rm -f {}".format(imname), shell=True)

    return output_fits


def search_vlass(c, crop=False, crop_scale=256, consider_QA_rejected=False):
    """ 
    Searches the VLASS catalog for a source

    Parameters
    ----------
    c: coordinates as SkyCoord object
    """
    # Find the VLASS tile
    tiles = get_tiles()
    tilename, epoch, obsdate = search_tiles(tiles, c)

    subtiles, c_tiles = get_subtiles(tilename, epoch, consider_QA_rejected)
    dist = c.separation(c_tiles)
    subtile = subtiles[np.argmin(dist)]

    imname = "{}.I.iter1.image.pbcor.tt0.subim.fits".format(subtile[:-1])
    if len(glob.glob(imname)) == 0:
        if epoch[5] == "1":
            url_get = "https://vlass-dl.nrao.edu/vlass/quicklook/{}v2/{}/{}".format(epoch, tilename, subtile)
        else:
            url_get = "https://vlass-dl.nrao.edu/vlass/quicklook/{}/{}/{}".format(epoch, tilename, subtile)
        fname = "{}{}".format(url_get, imname)
        subprocess.call("wget {}".format(fname), shell=True)
        if not os.path.exists(subtile) and consider_QA_rejected:
            if epoch[5] == "1":
                url_get = "https://vlass-dl.nrao.edu/vlass/quicklook/{}v2/QA_REJECTED/{}".format(epoch, subtile)
            else:
                url_get = "https://vlass-dl.nrao.edu/vlass/quicklook/{}/QA_REJECTED/{}".format(epoch, subtile)
            fname = "{}{}".format(url_get, imname)
            subprocess.call("wget {}".format(fname), shell=True)
    if crop:    
        out = get_cutout(imname, c, crop_scale=crop_scale)
        return out
    else:
        return imname


if __name__=="__main__":
    parser = argparse.ArgumentParser(description=\
        '''
        Searches VLASS for a source.
        User needs to supply RA (in decimal degrees), Dec (in decimal degrees)
        
        Usage: vlass_search.py <RA [deg]> <Dec [deg]> or vlass_search.py <MS>
        ''', formatter_class=argparse.RawTextHelpFormatter)
    
    #Test coords: 232.517541667 60.4054444
    try:
        ra = 15*(float(sys.argv[1]) + float(sys.argv[2])/60 + float(sys.argv[3])/3600)
        dec = float(sys.argv[4]) + sign(float(sys.argv[4]))*float(sys.argv[5])/60 + sign(float(sys.argv[4]))*float(sys.argv[6])/3600
    except IndexError:
        ra = float(sys.argv[1])
        dec = float(sys.argv[2])
    except ValueError:
        pi = 3.14159265358979
        import casacore.tables as pt
        table = pt.table(sys.argv[1]+"::FIELD")
        direction = table.getcol('PHASE_DIR').squeeze()
        ra=(direction[0]%(2*pi))/pi*180
        dec=direction[1]/pi*180
        table.close()
    except Exception:
        raise TypeError("Incorrect inputs given. Usage: vlass_search.py <RA [deg]> <Dec [deg]> or vlass_search.py <MS>")
        
    c = SkyCoord(ra, dec, unit='deg')

    if not glob.glob(summary_file_location):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), summary_file_location)
    
    search_vlass(c, crop=crop, crop_scale=crop_scale, consider_QA_rejected=consider_QA_rejected) 


import numpy 


def main():
    from plot_field.plot_field import main as plot_field_main

    plot_field_main()


def hello(n: int) -> str:
    """Greet the sum from 0 to n (exclusive end)."""
    sum_n = numpy.arange(n).sum()
    return f"Hello {sum_n}!"


#!/usr/bin/env python
from __future__ import print_function
import os
import logging
import io
import numpy as np
import pyvo as vo
from astropy.table import Table, Column, hstack, unique
import argparse

# from lofarpipe.support.data_map import DataMap
# from lofarpipe.support.data_map import DataProduct
import requests
from astropy.coordinates import SkyCoord
from astropy import units as u
from astropy.io import fits
from astropy.wcs import WCS
from requests.adapters import Retry, HTTPAdapter
from time import sleep

def format_skycoord(skycoord: SkyCoord) -> str:
    """
    Convert an Astropy SkyCoord to a string formatted as "hhmmss.ss+ddmmss.s"

    Arguments:
        skycoord (astropy.coordinates.SkyCoord): Input coordinate to format.

    Returns:
        str: Formatted string in the format "hhmmss.ss+ddmmss.s"
    """
    ra = skycoord.ra.to_value('hourangle')
    dec = skycoord.dec.to_value('deg')

    ra_hours = int(ra)
    ra_minutes = int((ra - ra_hours) * 60)
    ra_seconds = ((ra - ra_hours) * 60 - ra_minutes) * 60

    dec_degrees = int(dec)
    dec_arcminutes = int(abs((dec - dec_degrees) * 60))
    dec_arcseconds = abs(((dec - dec_degrees) * 60 - dec_arcminutes) * 60)

    ra_str = f"ILTJ{ra_hours:02d}{ra_minutes:02d}{ra_seconds:05.2f}"

    dec_sign = '+' if dec >= 0 else '-'
    dec_degrees = abs(dec_degrees)
    dec_str = f"{dec_sign}{dec_degrees:02d}{dec_arcminutes:02d}{dec_arcseconds:04.1f}"

    return ra_str + dec_str

def sum_digits(n):
    s = 0
    c = 0
    for ii in n:
        if ii != "-":
            s += int(ii)
            c += 1
    norm = float(s) / float(c)
    return norm


def count_p(n):
    s = 0
    c = 0
    for ii in n:
        if ii != "-":
            c += 1
            if ii == "P":
                s += 1
    norm = float(s) / float(c)
    return norm


def count_s(n):
    s = 0
    c = 0
    for ii in n:
        if ii != "-":
            c += 1
            if ii == "S":
                s += 1
    norm = float(s) / float(c)
    return norm


def count_x(n):
    s = 0
    c = 0
    for ii in n:
        if ii != "-":
            c += 1
            if ii == "X":
                s += 1
    norm = float(s) / float(c)
    return norm


def calculate_gval(goodness):
    """Calculate gval for Frits' calibrator metric"""
    return sum([int(score) for score in goodness if score != "-"]) / len(
        goodness.replace("-", "")
    )


def grab_coo_MS(MS):
    """
    Read the coordinates of a field from one MS corresponding to the selection given in the parameters
    Parameters
    ----------
    MS : str
        Full name (with path) to one MS of the field
    Returns
    -------
    RA, Dec : float,float
        coordinates of the field (RA, Dec in deg , J2000)
    """
    import casacore.tables as pt

    # reading the coordinates ("position") from the MS
    # NB: they are given in rad,rad (J2000)
    [[[ra, dec]]] = pt.table(MS + "/FIELD", readonly=True, ack=False).getcol(
        "PHASE_DIR"
    )
    # [[[ra,dec]]] = pt.table(MS, readonly=True, ack=False).getcol('PHASE_DIR')

    # RA is stocked in the MS in [-pi;pi]
    # => shift for the negative angles before the conversion to deg (so that RA in [0;2pi])
    if ra < 0:
        ra = ra + 2 * np.pi

    # convert radians to degrees
    ra_deg = ra / np.pi * 180.0
    dec_deg = dec / np.pi * 180.0

    # and sending the coordinates in deg
    return ra_deg, dec_deg


def input2strlist_nomapfile(invar):
    """from bin/download_IONEX.py
    give the list of MSs from the list provided as a string
    """

    str_list = None
    if type(invar) is str:
        if invar.startswith("[") and invar.endswith("]"):
            str_list = [f.strip(" '\"") for f in invar.strip("[]").split(",")]
        else:
            str_list = [invar.strip(" '\"")]
    elif type(invar) is list:
        str_list = [str(f).strip(" '\"") for f in invar]
    else:
        raise TypeError("input2strlist: Type " + str(type(invar)) + " unknown!")
    return str_list


def my_lotss_catalogue(
    RATar,
    DECTar,
    Radius=1.5,
    bright_limit_Jy=5.0,
    faint_limit_Jy=0.0,
    outfile="",
    use_vo=False,
):
    """
    Download the LoTSS skymodel for the target field
    Parameters
    ----------
    ms_input : str
        String from the list (map) of the target MSs
    Radius : float (default = 1.5)
        Radius for the LOTSS cone search in degrees

    """
    ## first check if the file already exists
    if os.path.isfile(outfile):
        print("LOTSS Skymodel for the target field exists on disk, reading in.")
        tb_final = Table.read(outfile, format="csv")
        if "Source_Name" in tb_final.colnames:
            tb_final.rename_column("Source_Name", "Source_id")
            tb_final.write(outfile, format="csv")
    else:
        print("DOWNLOADING LOTSS Skymodel for the target field")
        print("Radius is", Radius)
        # Reading a MS to find the coordinate (pyrap)
        # RATar, DECTar = grab_coo_MS(input2strlist_nomapfile(ms_input)[0])

        if use_vo:
            ## this is the tier 1 database to query
            # url = 'http://vo.astron.nl/lofartier1/q/cone/scs.xml'
            # HETDEX database.
            # url = 'https://vo.astron.nl/hetdex/lotss-dr1/cone/scs.xml'
            url = "https://vo.astron.nl/lotss_dr2/q/src_cone/scs.xml"

            ## query the database
            query = vo.dal.scs.SCSQuery(url, maxrec=10000000)
            query["RA"] = float(RATar)
            query["DEC"] = float(DECTar)
            query.radius = float(Radius)
            t = query.execute()

            ## convert to VO table
            try:
                tb = t.votable.to_table()
            except AttributeError:
                # Above statement didn't work, try the alternative.
                tb = t.to_table()
        else:
            # To get the combined source catalogue, query the surveys server
            print("Using the lofar-surveys catalogue!")
            session = requests.Session()
            # Will wait for 0, 20, 40 seconds between attempts.
            retries = Retry(
                total=8, backoff_factor=0.2, status_forcelist=[500, 502, 503, 504],
                allowed_methods=["GET", "POST"]
            )
            session.mount("https://", HTTPAdapter(max_retries=retries))

            r = session.get(
                "https://lofar-surveys.org/catalogue_search.csv?ra=%f&dec=%f&radius=%f"
                % (RATar, DECTar, Radius),
                timeout=50,
            )
            # Successful HTTP requests return 200.
            if r.status_code != 200:
                raise RuntimeError(
                    "Unsuccessful HTTP request querying https://lofar-surveys.org/catalogue_search.csv"
                )
            tb = Table.read(r.text.split("\n"), format="csv")
            print("Got a table with", len(tb), "entries")
            tb["Maj"].name = "Majax"
            tb["Min"].name = "Minax"
            session.close()

        flux_sort = tb.argsort("Total_flux")
        tb_sorted = tb[flux_sort[::-1]]
        ## and keep only some of the columns
        colnames = tb_sorted.colnames
        ## first check for a resolved column
        if "Resolved" not in colnames:
            resolved = np.where(
                is_resolved(
                    tb_sorted["Total_flux"],
                    tb_sorted["Peak_flux"],
                    tb_sorted["Isl_rms"],
                ),
                "R",
                "U",
            )
            tb_sorted["Resolved"] = resolved
        ## rename source_id column if necessary
        if "Source_Name" in tb_sorted.colnames:
            tb_sorted.rename_column("Source_Name", "Source_id")
        keep_cols = [
            "Source_id",
            "RA",
            "DEC",
            "Total_flux",
            "Peak_flux",
            "Majax",
            "Minax",
            "PA",
            "DC_Maj",
            "DC_Min",
            "DC_PA",
            "Isl_rms",
            "Resolved",
        ]
        if "LGZ_Size" in colnames:
            keep_cols = keep_cols + ["LGZ_Size", "LGZ_Width", "LGZ_PA"]
        if "LAS" in colnames:
            keep_cols = keep_cols + ["LAS"]

        tb_final = tb_sorted[keep_cols]

        tb_final = tb_final[tb_final["Total_flux"] >= faint_limit_Jy * 1e3]

        tb_final = tb_final[tb_final["Total_flux"] <= bright_limit_Jy * 1e3]

        tb_final.write(outfile, format="csv")

    return tb_final


def my_lbcs_catalogue(RATar, DECTar, Radius=1.5, outfile=""):
    """
    Download the LBCS skymodel for the target field
    Parameters
    ----------
    ms_input : str
        String from the list (map) of the target MSs
    Radius : float (default = 1.5)
        Radius for the LOTSS cone search in degrees

    """
    ## first check if the file already exists
    # print( outfile )
    if os.path.isfile(outfile):
        print("LBCS Skymodel for the target field exists on disk, reading in.")
        tb = Table.read(outfile, format="csv")
        return tb
    else:
        print("DOWNLOADING LBCS Skymodel for the target field")

        # Reading a MS to find the coordinate (pyrap)
        # RATar, DECTar = grab_coo_MS(input2strlist_nomapfile(ms_input)[0])

    ## construct an html query and try to connect
    # print(RATar, DECTar)
    url = "https://lofar-surveys.org/lbcs-search.fits?ra=%f&dec=%f&radius=%f" % (
        float(RATar),
        float(DECTar),
        float(Radius),
    )
    connected = False
    while not connected:
        try:
            response = requests.get(url, stream=True, verify=True, timeout=60)
            if response.status_code != 200:
                print(response.headers)
                raise RuntimeError("Code was %i" % response.status_code)
        except requests.exceptions.ConnectionError:
            print("Connection error! sleeping 30 seconds before retry...")
            sleep(30)
        except (requests.exceptions.Timeout, requests.exceptions.ReadTimeout):
            print("Timeout! sleeping 30 seconds before retry...")
            sleep(30)
        else:
            connected = True

    mem = io.BytesIO()
    for chunk in response.iter_content(chunk_size=8192):
        if chunk:
            mem.write(chunk)
    mem.seek(0)
    tb = Table.read(mem)
    mem.close()
    del response

    # print( len( tb ) )

    ### Workaround to sort and pick good calibrator info from tb array ###########

    if not len(tb) > 1:
        logging.critical(
            "There are no LBCS sources within the given radius. Check your source is within the LBCS footprint and increase the search radius. Exiting..."
        )
        return
    else:
        ## calculate the total FT goodness
        ft_total = []
        for xx in range(len(tb)):
            ft_total.append(sum_digits(tb[xx]["FT_Goodness"]))
        ft_col = Column(ft_total, name="FT_total")
        tb.add_column(ft_col)

    tb.write(outfile, format="csv")

    return tb


def find_close_objs(lo, lbcs, tolerance=5.0):
    ## first filter the LBCS data on Flags
    lbcs_idx = np.where(np.logical_or(lbcs["Flags"] == "O", lbcs["Flags"] == "A"))
    lbcs = lbcs[lbcs_idx]
    ## get rid of anything with only X's
    nump = []
    nums = []
    for xx in range(len(lbcs)):
        nump.append(count_p(lbcs["Goodness"][xx]))
        nums.append(count_s(lbcs["Goodness"][xx]))
    # print( np.array(nump) + np.array(nums) )
    lbcs_idx = np.where(np.array(nump) + np.array(nums) > 0)[0]
    lbcs = lbcs[lbcs_idx]

    ## get RA and DEC for the catalogues
    lotss_coords = SkyCoord(lo["RA"], lo["DEC"], frame="icrs", unit="deg")
    lbcs_coords = SkyCoord(lbcs["RA"], lbcs["DEC"], frame="icrs", unit="deg")

    ## search radius
    search_rad = tolerance / 60.0 / 60.0 * u.deg

    ## loop through the lbcs coordinates -- this will be much faster than looping through lotss
    lotss_idx = []
    lbcs_idx = []
    for xx in range(len(lbcs)):
        seps = lbcs_coords[xx].separation(lotss_coords)
        match_idx = np.where(seps < search_rad)[0]
        if len(match_idx) == 0:
            # there's no match, move on to the next source
            m_idx = [-1]
            pass
        else:
            if len(match_idx) == 1:
                ## there's only one match
                m_idx = match_idx[0]
                lbcs_idx.append(xx)
                lotss_idx.append(m_idx)
            if len(match_idx) > 1:
                ## there's more than one match, pick the brightest
                tmp = lo[match_idx]
                m_idx = np.where(tmp["Total_flux"] == np.max(tmp["Total_flux"]))[0]
                if not isinstance(m_idx, int):
                    m_idx = m_idx[0]
                lbcs_idx.append(xx)
                lotss_idx.append(m_idx)
    lbcs_matches = lbcs[lbcs_idx]
    lotss_matches = lo[lotss_idx]

    combined = hstack([lbcs_matches, lotss_matches], join_type="exact")
    ## check if there are duplicate lbcs observations for a lotss source
    if len(np.unique(combined["Source_id"])) != len(combined):
        # there are duplicates
        print("There are duplicate LBCS sources, selecting the best candidate(s).")
        src_ids = np.unique(combined["Source_id"])
        good_idx = []
        for src_id in src_ids:
            idx = np.where(combined["Source_id"] == src_id)[0]
            if len(idx) > 1:
                ## multiple matches found.  Count P's first and then break ties with Goodness_FT
                num_P = []
                total_ft = []
                for yy in range(len(idx)):
                    tmp = combined[idx[yy]]["Goodness"]
                    num_P.append(count_p(tmp))

                    tmp = combined[idx[yy]]["FT_Goodness"]
                    total_ft.append(sum_digits(tmp))
                ## check that the total_ft values are non-zero before looking for a best cal
                if np.max(total_ft) > 0:
                    ## pick the one with the highest number of P's -- if tie, use total_ft
                    best_idx = np.where(num_P == np.max(num_P))[0]  ## this is an array
                if len(best_idx) == 1:
                    good_idx.append(idx[best_idx][0])  ## idx[best_idx][0] is a number
                if len(best_idx) > 1:
                    currentmax = 0.0
                    for i in range(0, len(best_idx)):
                        if total_ft[best_idx[i]] > currentmax:
                            currentmax = total_ft[best_idx[i]]
                            ft_idx = i
                    good_idx.append(idx[best_idx[ft_idx]])
                else:
                    print("Duplicate sources have total_ft = 0, removing from results.")
            else:
                good_idx.append(idx[0])

        result = combined[good_idx]
    else:
        print("No duplicate sources found")
        result = combined
    ## rename RA columns
    result.rename_column("RA_1", "RA_LBCS")
    result.rename_column("DEC_1", "DEC_LBCS")
    result.rename_column("RA_2", "RA")
    result.rename_column("DEC_2", "DEC")

    return result


def is_resolved(Sint, Speak, rms):
    """Determines if a source is resolved or unresolved.
    The calculation is presented in Shimwell et al. 2018 of the LOFAR DR1 paper splash.

    Args:
        Sint (float or ndarray): integrated flux density.
        Speak (float or ndarray): peak flux density.
        rms (float or ndarray): local rms around the source.
    Returns:
        resolved (bool or ndarray): True if the source is resolved, False if not.
    """
    resolved = (Sint / Speak) > 1.25 + 3.1 * (Speak / rms) ** (-0.53)
    return resolved


def remove_multiples_position(mycat, racol="RA", decol="DEC"):
    radecstrings = []
    for i in np.arange(0, len(mycat)):
        radecstrings.append(str(mycat[i][racol]) + str(mycat[i][decol]))
    radecstrings = np.asarray(radecstrings)
    if len(np.unique(radecstrings)) != len(mycat):
        radecs = np.unique(radecstrings)
        good_idx = []
        for radec in radecs:
            idx = np.where(radecstrings == radec)[0]
            if len(idx) > 1:
                ## multiple matches found.  Count P's first and then break ties with Goodness_FT
                num_P = []
                total_ft = []
                for yy in range(len(idx)):
                    tmp = mycat[idx[yy]]["Goodness"]
                    num_P.append(count_p(tmp))
                    tmp = mycat[idx[yy]]["FT_Goodness"]
                    total_ft.append(sum_digits(tmp))
                ## check that the total_ft values are non-zero before looking for a best cal
                if np.max(total_ft) > 0:
                    ## pick the one with the highest number of P's -- if tie, use total_ft
                    best_idx = np.where(num_P == np.max(num_P))[0]  ## this is an array
                    if len(best_idx) == 1:
                        good_idx.append(
                            idx[best_idx][0]
                        )  ## idx[best_idx][0] is a number
                    if len(best_idx) > 1:
                        currentmax = 0.0
                        for i in range(0, len(best_idx)):
                            if total_ft[best_idx[i]] > currentmax:
                                currentmax = total_ft[best_idx[i]]
                                ft_idx = i
                        good_idx.append(idx[best_idx[ft_idx]])
                else:
                    print("Duplicate sources have total_ft = 0, removing from results.")
            else:
                good_idx.append(idx[0])
        mycat = mycat[good_idx]
    else:
        print("All LBCS sources are unique")

    return mycat


def mkfits(rasiz, decsiz, imsiz, pixsiz):
    hdu = fits.PrimaryHDU(np.zeros((int(imsiz), int(imsiz))))
    hdu.header.update({"CTYPE1": "RA---SIN"})
    hdu.header.update({"CRVAL1": rasiz})
    hdu.header.update({"CRPIX1": imsiz / 2.0})
    hdu.header.update({"CDELT1": -pixsiz})
    hdu.header.update({"CTYPE2": "DEC--SIN"})
    hdu.header.update({"CRVAL2": decsiz})
    hdu.header.update({"CRPIX2": imsiz / 2.0})
    hdu.header.update({"CDELT2": pixsiz})
    hdu.header.update({"EQUINOX": 2000.0})
    hdu.header.update({"EPOCH": 2000.0})
    #    hdu.data = np.random.random(imsiz*imsiz).reshape(imsiz,imsiz)
    if os.path.exists("temp.fits"):
        os.system("rm temp.fits")
    hdu.writeto("temp.fits")


def smearing_calculation(
    nchan=16,
    obs_freq=144000000,
    radii=np.arange(0.001, 4, 0.00001),
    resolution=0.3 / 206265 * 180 / np.pi,
    av_time=1,
    thresholds=[0.2, 0.4, 0.6, 0.8],
):
    from scipy.special import erf

    bandwidth = 1.95e3 / nchan

    beta_fac = bandwidth / obs_freq * radii / resolution
    gamma_fac = 2.0 * np.sqrt(np.log(2.0))
    reduction_bandwidth = (
        np.sqrt(np.pi) / (gamma_fac * beta_fac) * erf(gamma_fac * beta_fac / 2)
    )
    reduction_time = 1 - 1.22e-9 * (radii / resolution) ** 2 * av_time**2
    reduction_total = reduction_bandwidth * reduction_time
    # print(reduction_bandwidth, reduction_total)
    thres_radii = []
    for thresh in thresholds:
        idx = np.argwhere(reduction_total < thresh)[0]
        # print(radii[idx])
        thres_radii.append(radii[idx])
    return thres_radii


def angular_distance(RA1, DEC1, RA2, DEC2):
    c1 = SkyCoord(RA1, DEC1, unit="deg")
    c2 = SkyCoord(RA2, DEC2, unit="deg")
    return c1.separation(c2).value


def smallest_distance(RA, DEC, lbcs_catalogue):
    distances = [
        angular_distance(RA, DEC, source["RA"], source["DEC"])
        for source in lbcs_catalogue
    ]

    ids = np.argmin(distances)

    return ids, distances


def make_plot(
    RA,
    DEC,
    lotss_catalogue,
    extreme_catalogue,
    lbcs_catalogue,
    targRA=None,
    targDEC=None,
    nchan=16,
    av_time=1,
    outdir=".",
):
    import matplotlib.pyplot as plt
    from astropy.visualization.wcsaxes import SphericalCircle

    mkfits(RA, DEC, 2048, 2.6367)

    f = fits.open("temp.fits")
    w = WCS(f[0].header)

    lbcs = lbcs_catalogue

    if os.path.exists(lotss_catalogue):
        lotss = Table.read(lotss_catalogue)
        avg_flux = np.median(lotss["Total_flux"])
    scaling = 0.1

    thres_radii = smearing_calculation(nchan=nchan, av_time=av_time)

    plt.figure(figsize=(12, 12))

    ax = plt.subplot(
        projection=w,
    )

    color = 0.5
    fraction = ["80%", "60%", "40%", "20%"]
    for i, thresh in enumerate(thres_radii):
        color = color + 0.1
        centre_coord = SkyCoord(RA, DEC, unit="deg")
        c = SphericalCircle(
            centre_coord,
            thresh * u.deg,
            edgecolor=None,
            facecolor=str(color),
            transform=ax.get_transform("fk5"),
            zorder=-1,
        )
        # print(RA, RA + thresh[0])
        # ax.text(
        #     RA + thresh[0],
        #     DEC,
        #     fraction[i],
        #     transform=ax.get_transform("fk5"),
        #     fontsize="large",
        # )
        ax.add_patch(c)

    ax.scatter(
        RA, DEC, marker="x", s=30, transform=ax.get_transform("fk5"), label="Ptg Centre"
    )

    if os.path.exists(lotss_catalogue):
        ax.scatter(
            lotss["RA"],
            lotss["DEC"],
            transform=ax.get_transform("fk5"),
            s=lotss["Total_flux"] * scaling,
            label="Potential Targets",
        )

    if len(extreme_catalogue) > 0:
        for source in extreme_catalogue:
            vector_orig = [source["RA"] - RA, source["DEC"] - DEC]
            norm = np.sqrt(vector_orig[0] ** 2 + vector_orig[1] ** 2)
            vector = [vector_orig[0] / norm, vector_orig[1] / norm]
            alt = [vector[0] - RA, vector[1] - DEC]
            print(vector, alt)
            ax.arrow(
                RA + 2 * vector[0],
                DEC + 2 * vector[1],
                0.5 * vector[0],
                0.5 * vector[1],
                transform=ax.get_transform("fk5"),
                width=0.01,
                head_width=0.05,
                head_length=0.05,
                length_includes_head=True,
            )
            ax.text(
                RA + 2.5 * vector[0],
                DEC + 2.5 * vector[1],
                s="%.2f" % (source["Total_flux"] / 1000) + " Jy - %.2f degrees" % norm,
                transform=ax.get_transform("fk5"),
            )

    ax.scatter(
        lbcs["RA"],
        lbcs["DEC"],
        transform=ax.get_transform("fk5"),
        s=60,
        label="LBCS Sources",
    )

    c = SphericalCircle(
        centre_coord,
        1.5 * u.deg,
        edgecolor="yellow",
        facecolor="none",
        transform=ax.get_transform("fk5"),
    )
    ax.add_patch(c)

    ### Calculate and plot smallest distance to target
    dist_ids, dist = smallest_distance(RA, DEC, lbcs)

    closest_calib = lbcs[dist_ids]
    ax.plot(
        [RA, lbcs_catalogue[0]["RA"]],
        [DEC, lbcs_catalogue[0]["DEC"]],
        linestyle="--",
        linewidth=2,
        transform=ax.get_transform("fk5"),
        label="Distance = %.2f degrees" % dist[0],
    )

    if targRA != None:
        ax.scatter(
            targRA,
            targDEC,
            marker="x",
            s=80,
            transform=ax.get_transform("fk5"),
            label="Target",
        )
        dist = angular_distance(
            targRA, targDEC, lbcs_catalogue[0]["RA"], lbcs_catalogue[0]["DEC"]
        )

        ax.plot(
            [targRA, lbcs_catalogue[0]["RA"]],
            [targDEC, lbcs_catalogue[0]["DEC"]],
            linestyle="--",
            color="black",
            linewidth=2,
            transform=ax.get_transform("fk5"),
            label="Distance = %.2f degrees" % dist,
        )

    for i in range(len(lbcs)):
        ax.text(
            lbcs[i]["RA"] - 0.05,
            lbcs[i]["DEC"],
            "%.2f Jy" % (lbcs[i]["Total_flux"] / 1000),
            transform=ax.get_transform("fk5"),
        )
    plt.legend(fontsize="large", loc="upper right")

    plt.savefig(os.path.join(outdir, f"{RA}_{DEC}_output.png"))
    os.system("rm temp.fits")


def convert_vlass_fits(fitsfile):
    import matplotlib.pyplot as plt
    from astropy.visualization import PercentileInterval, imshow_norm
    from astropy.wcs import WCS

    print("Processing %s" % fitsfile)
    header = fits.open(fitsfile)[0].header
    wcs = WCS(header).celestial  # Ignore frequency/stokes axis

    image_data = fits.getdata(fitsfile)

    # Shape is (1,1, 3722, 3722). Plot the first image
    interval = PercentileInterval(99.9)
    process_data = interval(image_data)
    plt.subplot(projection=wcs)
    imshow_norm(process_data, cmap="gray")

    # Axis Labels
    plt.xlabel("RA")
    plt.ylabel("Dec")

    plt.savefig(fitsfile[:-5] + ".png")


def convert_cutout(fitsfile):
    import matplotlib.pyplot as plt
    from astropy.visualization import PercentileInterval, imshow_norm
    from astropy.wcs import WCS

    print("Processing %s" % fitsfile)
    header = fits.open(fitsfile)[0].header
    wcs = WCS(header).celestial  # Ignore frequency/stokes axis

    image_data = fits.getdata(fitsfile)

    # Shape is (1,1, 3722, 3722). Plot the first image
    interval = PercentileInterval(99.9)
    process_data = interval(image_data)
    plt.subplot(projection=wcs)
    imshow_norm(process_data, cmap="Blues")

    # Remove all axes
    plt.axis("off")

    plt.savefig("assets/cutout.png", bbox_inches="tight", pad_inches=0)


def fit_spectrum(delay_cals_file, outdir):
    from plot_field.fit_synchrotron_spectrum import (
            fit_from_NED,
            fit_from_trusted_surveys,
        )
    result = Table.read(delay_cals_file)
    # Empty columns for fitting parameters within delay calibration
    total_flux_column = Column([None] * len(result), name="fit_flux", unit="Jy")
    alpha_1_column = Column([None] * len(result), name="alpha_1")
    alpha_2_column = Column([None] * len(result), name="alpha_2")
    survey_column = Column([None] * len(result), name="Catalogue")
    chi_sqr = Column([None] * len(result), name="chi_sqr")
    no_points_column = Column([None] * len(result), name="phot_points")
    result.add_column(total_flux_column)
    result.add_column(alpha_1_column)
    result.add_column(alpha_2_column)
    result.add_column(chi_sqr)
    result.add_column(no_points_column)
    result.add_column(survey_column)
    result.write(delay_cals_file, format="csv", overwrite=True)

    for i, source in enumerate(result):
        fit_parameters_trusted = None
        fit_parameters_NED = None
        chi_square = None
        no_points = 0

        try:
            fit = fit_from_trusted_surveys(source["RA"], source["DEC"], 12.0, outdir)
            fit_parameters_trusted, chi_square = fit.fit_parameters, fit.stats
            no_points = fit.freq_points
            if chi_square > 20 or np.abs(fit_parameters_trusted[1]) > 3:
                print(
                    f"Calibrator {source['Source_id']} has a poor fit! (χ² = {chi_square:.2f})"
                )
                fit_parameters_trusted, chi_square = None, None
                raise ValueError
        except (ValueError, TypeError, IndexError):
            try:
                fit = fit_from_NED(source["RA"], source["DEC"], 12.0, outdir)
                fit_parameters_NED, chi_square = fit.fit_parameters, fit.stats
                no_points = fit.freq_points
            except (ValueError, TypeError):
                print(
                    f"Calibrator {source['Source_id']} could not be fit with NED or trusted surveys!"
                )

        if fit_parameters_NED is not None:
            fitting_parameters = np.append(fit_parameters_NED, "NED")
        elif fit_parameters_trusted is not None:
            fitting_parameters = np.append(fit_parameters_trusted, "Trusted")
        else:
            fitting_parameters = [None, None, None, None]

        result["fit_flux"][i] = fitting_parameters[0]
        result["alpha_1"][i] = fitting_parameters[1]
        result["alpha_2"][i] = fitting_parameters[2]
        result["chi_sqr"][i] = chi_square
        result["phot_points"][i] = no_points
        result["Catalogue"][i] = fitting_parameters[3]
    result.write(delay_cals_file, format="csv", overwrite=True)


def ps_match(file):
    t = Table.read(file)

    ps_id_column = Column([None] * len(t), name="ps_id")
    ps_ra_column = Column([None] * len(t), name="ps_RA", unit="deg")
    ps_dec_column = Column([None] * len(t), name="ps_DEC", unit="deg")

    t.add_column(ps_id_column)
    t.add_column(ps_ra_column)
    t.add_column(ps_dec_column)

    targets = []
    for source in t:
        id = source["Source_id"]
        ra = source["RA"]
        dec = source["DEC"]

        dict = {"ra": str(ra), "dec": str(dec), "target": str(id)}

        targets.append(dict)

    data = {
        "targets": targets,
        "resolve": "False",
        "radius": 0.000833 / 3,
    }  # 1 arcsecond

    URL = "https://catalogs.mast.stsci.edu/api/v0.1/panstarrs/dr2/mean/crossmatch/csv"
    r = requests.post(URL, json=data)
    if len(r.text) > 10:
        out_t = Table.read(r.text, format="csv")

        out_t = unique(out_t, keys="_searchID_")
        out_t.sort(keys=["_searchID_"])

        for source in out_t:
            index = source["_searchID_"]
            t["ps_id"][index] = source["MatchID"]
            t["ps_RA"][index] = source["MatchRA"]
            t["ps_DEC"][index] = source["MatchDEC"]

    t.write(file, overwrite=True)


def gaia_quasar_match(file):
    from astroquery.vizier import Vizier

    t = Table.read(file)
    t["_RAJ2000"] = t["RA"] * u.deg
    t["_DEJ2000"] = t["DEC"] * u.deg
    CatNorth = "J/ApJS/271/54/table4"

    v = Vizier(catalog=CatNorth, columns=["Gaia", "RA_ICRS", "DE_ICRS"])

    # Match against table
    print(f"Querying GAIA quasars for {file}")
    out = v.query_region(t, radius="2s", inner_radius="0.01s", cache=False)
    if out:
        key = out.keys()[0]
        print(f"Found {len(out[key])} quasar cross-matches")
    else:
        print("Found no cross matches to GAIA quasars")

    gaia_id_column = Column([None] * len(t), name="gaia_id")
    gaia_ra_column = Column([None] * len(t), name="gaia_RA", unit="deg")
    gaia_dec_column = Column([None] * len(t), name="gaia_DEC", unit="deg")

    t.add_column(gaia_id_column)
    t.add_column(gaia_ra_column)
    t.add_column(gaia_dec_column)

    for source in out:
        index = source["_q"] - 1
        t["gaia_id"][index] = source["Gaia"]
        t["gaia_RA"][index] = source["RA_ICRS"]
        t["gaia_DEC"][index] = source["DE_ICRS"]

    t.write(file, overwrite=True)


def make_html(
    RATar,
    DECTar,
    lotss_result_file,
    extreme_catalogue,
    result,
    targRA,
    targDEC,
    nchan,
    av_time,
    pointing,
):
    # Check if required packages are installed
    try:
        from dash import Dash, dcc, html, Input, Output
        import plotly.graph_objects as go
        from dash.exceptions import PreventUpdate
    except ImportError:
        # If not, inform user of commands to install
        print("Please install the following packages to run this function:")
        print("dash, plotly, pandas")
        print("You can install them by running the following command:")
        print("pip install dash plotly pandas")

    # Copy all images to an assets folder
    if os.path.exists("assets"):
        pass
    else:
        os.mkdir("assets")
        os.system("cp *.png assets")

    # Small molcule drugbank dataset
    # Source: https://raw.githubusercontent.com/plotly/dash-sample-apps/main/apps/dash-drug-discovery/data/small_molecule_drugbank.csv'
    data_path = "delay_calibrators.csv"

    df = Table.read(data_path, format="csv")
    #    df = pd.read_csv(data_path, header=0,)

    print(df)

    fig = go.Figure(
        data=[
            go.Scatter(
                x=df["RA"],
                y=df["DEC"],
                mode="markers",
                customdata=df["Observation"],
                marker=dict(
                    colorscale="viridis",
                    color=df["Total_flux"],
                    size=20,
                    colorbar={"title": "Total<br>Flux[mJy]"},
                    line={"color": "#444"},
                    reversescale=True,
                    sizeref=45,
                    sizemode="diameter",
                    opacity=0.8,
                ),
            )
        ]
    )

    # Add image_catalogue scatter
    # On click, display catalogue info above the plot
    df_lotss = Table.read(lotss_result_file, format="csv")

    # Remove sources close to the sources in df
    for i in range(len(df)):
        idx = np.where(
            angular_distance(df["RA"][i], df["DEC"][i], df_lotss["RA"], df_lotss["DEC"])
            < 0.001
        )[0]
        df_lotss.remove_rows(idx)

    fig.add_trace(
        go.Scatter(
            x=df_lotss["RA"],
            y=df_lotss["DEC"],
            mode="markers",
            customdata=df_lotss["Source_id"],
            marker=dict(
                color="red",
                sizeref=0.4,
                opacity=0.4,
                size=df_lotss["Total_flux"] / 100,
                line=dict(color="black", width=2),
            ),
            hoverinfo="text",
        )
    )

    # Add large circle indicating 1.5 degree field of view
    fig.add_shape(
        type="circle",
        xref="x",
        yref="y",
        x0=RATar - 1.5,
        y0=DECTar - 1.5,
        x1=RATar + 1.5,
        y1=DECTar + 1.5,
        line=dict(
            color="yellow",
        ),
    )

    # turn off native plotly.js hover effects - make sure to use
    # hoverinfo="none" rather than "skip" which also halts events.
    # fig.update_traces(hoverinfo="none", hovertemplate=None)
    # Square aspect ratio

    fig.update_layout(
        xaxis=dict(title="RA"),
        yaxis=dict(title="DEC"),
        plot_bgcolor="rgba(255,255,255,0.1)",
        showlegend=False,
        height=800,
        width=800,
    )

    # Update layout and update traces
    fig.update_layout(clickmode="event+select")

    app = Dash(__name__)

    # Add image of source to the layout
    # Add button which will display cutout on click
    app.layout = html.Div(
        [
            dcc.Graph(
                id="graph_interaction",
                figure=fig,
            ),
            html.Button("Display Cutout", id="cutout_button", n_clicks=0),
            html.Div(id="output"),
        ]
    )

    @app.callback(
        Output("output", "children"),
        [Input("graph_interaction", "clickData")],
    )
    def display_click_data(clickData):
        if clickData is None:
            raise PreventUpdate
        else:
            source_id = clickData["points"][0]["customdata"]
            if source_id[0] == "L":
                source = df[df["Observation"] == source_id]
                return html.Div(
                    [
                        html.Img(
                            src=app.get_asset_url("./" + source_id + "_vlass.png"),
                            style={"height": "400px", "width": "400px"},
                        ),
                        html.H2(str(source["Observation"])),
                        html.P("%f mJy" % source["Total_flux"]),
                        html.P(str(source["RA"])),
                        html.P(str(source["DEC"])),
                    ]
                )
            else:
                source = df_lotss[df_lotss["Source_id"] == source_id]
                return html.Div(
                    [
                        html.H2(str(source["Source_id"])),
                        html.P("%f mJy" % source["Total_flux"]),
                        html.P(str(source["RA"])),
                        html.P(str(source["DEC"])),
                    ]
                )

    @app.callback(
        Output("graph_interaction", "figure", allow_duplicate=True),
        Input("cutout_button", "n_clicks"),
        prevent_initial_call=True,
    )
    def display_cutout(n_clicks):
        if n_clicks % 2 == 0:
            # Remove image layer from the graph
            fig.layout.images = []
            return fig
        else:
            if os.path.exists("assets/cutout.png"):
                pass
            else:
                # Convert RA and DEC to equatorial
                cutout = (
                    "https://lofar-surveys.org/public/DR2/mosaics/"
                    + pointing
                    + "/low-mosaic-blanked.fits"
                )
                # Make request to get cutout. Direct file to assets folder
                response = requests.get(cutout)
                with open("assets/cutout.fits", "wb") as f:
                    f.write(response.content)

                convert_cutout("assets/cutout.fits")

            # Add to background of the graph. It should cover the entire graph
            fig.add_layout_image(
                dict(
                    source="assets/cutout.png",
                    xref="x",
                    yref="y",
                    x=RATar - 1.5,
                    y=DECTar + 1.5,
                    sizex=3,
                    sizey=3,
                    sizing="stretch",
                    layer="below",
                )
            )
        return fig

    return app


def generate_catalogues(
    RATar,
    DECTar,
    targRA=0.0,
    targDEC=0.0,
    lotss_radius=1.5,
    lbcs_radius=1.5,
    im_radius=1.24,
    bright_limit_Jy=5.0,
    lotss_catalogue="lotss_catalogue.csv",
    lbcs_catalogue="lbcs_catalogue.csv",
    lotss_result_file="image_catalogue.csv",
    delay_cals_file="delay_calibrators.csv",
    match_tolerance=5.0,
    image_limit_Jy=0.01,
    continue_no_lotss=False,
    nchan=16,
    av_time=1.0,
    vlass=False,
    html=False,
    outdir=".",
    pointing=None,
    fit_spec=False,
    force=False,
):
    # def plugin_main( RA, DEC, **kwargs ):
    #    im_radius = float(kwargs['im_radius'])
    #    image_limit_Jy = float(kwargs['image_limit_Jy'])
    #    bright_limit_Jy = float(kwargs['bright_limit_Jy'])
    #    match_tolerance = float(kwargs['match_tolerance'])
    # mslist = DataMap.load(mapfile_in)
    # MSname = mslist[0].file
    # For testing
    # MSname = kwargs['MSname']

    ## prepend everything with outdir
    if not os.path.exists(outdir):
        os.mkdir(outdir)
    ## if using a user-defined catalogue, don't rename it
    if lotss_catalogue == 'lotss_catalogue.csv':
        lotss_catalogue = os.path.join(outdir, lotss_catalogue)
    lbcs_catalogue = os.path.join(outdir, lbcs_catalogue)
    lotss_result_file = os.path.join(outdir, lotss_result_file)
    delay_cals_file = os.path.join(outdir, delay_cals_file)
    extreme_file = os.path.join(outdir, "extreme_catalogue.csv")

    ## first check for a valid delay_calibrator file
    if os.path.isfile(delay_cals_file):
    #if os.path.isfile(lbcs_catalogue) or os.path.isfile(lotss_catalogue):
        print("Delay calibrators file {:s} already exists!".format(delay_cals_file))
        if force:
            print("Forcing overwrite")
            os.remove(delay_cals_file)
            os.remove(lotss_catalogue)
            os.remove(lbcs_catalogue)
            try:
                os.remove(lotss_result_file)
                os.remove(extreme_file)
            except:
                print("No LoTSS files")
            pass
        else:
            print("Returning!")
            return

    ## look for or download LBCS
    print("Attempting to find or download LBCS catalogue.")
    lbcs_catalogue = my_lbcs_catalogue(
        RATar, DECTar, Radius=lbcs_radius, outfile=lbcs_catalogue
    )

    sleep(2)
    ## look for or download LoTSS
    print("Attempting to find or download LoTSS catalogue.")
    lotss_catalogue = my_lotss_catalogue(
        RATar,
        DECTar,
        Radius=lotss_radius,
        bright_limit_Jy=bright_limit_Jy,
        faint_limit_Jy=0.0,
        outfile=lotss_catalogue,
        use_vo=True,
    )

    print("Finding bright sources outside field")
    extreme_catalogue = my_lotss_catalogue(
        RATar,
        DECTar,
        Radius=5.0,
        bright_limit_Jy=1000.0,
        faint_limit_Jy=10.0,
        outfile=os.path.join(outdir, "extreme_catalogue.csv"),
        use_vo=True
    )
    extreme_catalogue = remove_multiples_position(extreme_catalogue)
    ## if lbcs exists, and either lotss exists or continue_without_lotss = True, continue
    ## else provide an error message and stop
    if len(lbcs_catalogue) == 0:
        logging.error("LBCS coverage does not exist, and catalogue not found on disk.")
        return
    if len(lotss_catalogue) == 0 and not continue_no_lotss:
        logging.error(
            "LoTSS coverage does not exist, and contine_no_lotss is set to False."
        )
        return

    ## if the LoTSS catalogue is empty, write out the delay cals only
    if len(lotss_catalogue) == 0 or continue_no_lotss:
        print(
            "Target field not in LoTSS coverage yet! Only writing {:s} based on LBCS".format(
                delay_cals_file
            )
        )

        ## Add the radius from phase centre to the catalogue
        # RATar, DECTar = grab_coo_MS(input2strlist_nomapfile(MSname)[0])
        ptg_coords = SkyCoord(RATar, DECTar, frame="icrs", unit="deg")

        src_coords = SkyCoord(
            lbcs_catalogue["RA"], lbcs_catalogue["DEC"], frame="icrs", unit="deg"
        )
        separations = src_coords.separation(ptg_coords)
        seps = Column(separations.deg, name="Radius")
        lbcs_catalogue.add_column(seps)

        # Create an ILTJhhmmss.ss+ddmmss.s name on-the-fly.
        lbcs_catalogue["Source_id"] = [format_skycoord(SkyCoord(ra, dec, unit="deg")) for (ra, dec) in zip(lbcs_catalogue["RA"], lbcs_catalogue["DEC"])]

        ## add in some dummy data
        Total_flux = Column(
            np.ones(len(lbcs_catalogue)) * 1e3, name="Total_flux", unit="mJy"
        )
        lbcs_catalogue.add_column(Total_flux)
        LGZ_Size = Column(
            np.ones(len(lbcs_catalogue)) * 20.0, name="LGZ_Size", unit="arcsec"
        )  ## set to a default of 20 arcsec
        lbcs_catalogue.add_column(LGZ_Size)

        ## remove duplicate sources if necessary
        lbcs_catalogue = remove_multiples_position(lbcs_catalogue)

        ## remove calibrators over 1.2 degrees from pointing centre
        lbcs_catalogue = lbcs_catalogue[lbcs_catalogue["Radius"] < 1.2]

        # Calculate the new score r/F
        goodness = [str(score) for score in lbcs_catalogue["FT_Goodness"]]
        print(goodness)
        g = [calculate_gval(calibrator) for calibrator in goodness]
        gscore = lbcs_catalogue["Radius"].value / g
        gscore_column = Column(gscore, name="gscore")
        lbcs_catalogue.add_column(gscore_column)

        ## order based on radius from the phase centre
        lbcs_catalogue.sort("gscore")

        ## order based on radius from the phase centre
        # lbcs_catalogue.sort('Radius')

        ## write the catalogue
        lbcs_catalogue.write(delay_cals_file, format="csv")

        result = lbcs_catalogue
    else:
        ## else continue
        result = find_close_objs(
            lotss_catalogue, lbcs_catalogue, tolerance=match_tolerance
        )
        ## check if there are any matches
        if len(result) == 0:
            logging.error(
                "LoTSS and LBCS coverage exists, but no matches found. This indicates something went wrong, please check your catalogues.\n Most likely LoTSS coverage only partially covers your pointing. You can re-run with --continue_no_lotss to ignore your LoTSS catalogue."
            )
            return
        else:
            # add radius to the catalogue
            # RATar, DECTar = grab_coo_MS(input2strlist_nomapfile(MSname)[0])
            # result = lbcs_catalogue
            ptg_coords = SkyCoord(RATar, DECTar, frame="icrs", unit="deg")

            src_coords = SkyCoord(result["RA"], result["DEC"], frame="icrs", unit="deg")
            separations = src_coords.separation(ptg_coords)
            seps = Column(separations.deg, name="Radius")
            result.add_column(seps)

            ## remove calibrators over 1.2 degrees from pointing centre
            result = result[result["Radius"] < 1.2]

            # Calculate the new score r/F
            goodness = [str(score) for score in result["FT_Goodness"]]
            print(goodness)
            g = [calculate_gval(calibrator) for calibrator in goodness]
            gscore = result["Radius"].value / g
            gscore_column = Column(gscore, name="gscore")
            result.add_column(gscore_column)

            ## order based on radius from the phase centre
            result.sort("gscore")

            # result.rename_column('Observation','Source_id')

            ## Write catalogues
            ## 1 - delay calibrators -- from lbcs_catalogue
            result.write(delay_cals_file, format="csv", overwrite=True)
            print(
                "Writing delay calibrator candidate file {:s}".format(delay_cals_file)
            )

            ## make a flux cut
            image_index = np.where(
                lotss_catalogue["Peak_flux"] >= image_limit_Jy * 1e3
            )[0]
            flux_cut_sources = lotss_catalogue[image_index]

            ## make a radius cut
            src_coords = SkyCoord(
                flux_cut_sources["RA"],
                flux_cut_sources["DEC"],
                frame="icrs",
                unit="deg",
            )
            separations = src_coords.separation(ptg_coords)
            seps = Column(separations.deg, name="Radius")
            flux_cut_sources.add_column(seps)
            good_idx = np.where(flux_cut_sources["Radius"] <= im_radius)[0]
            sources_to_image = flux_cut_sources[good_idx]

            nsrcs = float(len(sources_to_image))
            print(
                "There are "
                + str(len(result))
                + " delay calibrators within "
                + str(im_radius)
                + " degrees of the pointing centre"
            )
            print(
                "There are "
                + str(nsrcs)
                + " sources above "
                + str(image_limit_Jy * 1000)
                + " mJy within "
                + str(im_radius)
                + " degrees of the phase centre."
            )
            sources_to_image.write(lotss_result_file, format="csv", overwrite=True)
            gaia_quasar_match(lotss_result_file)
            ps_match(lotss_result_file)

    gaia_quasar_match(delay_cals_file)
    ps_match(delay_cals_file)

    print("Assumed averaging - nchannels: %s; time averaging: %s" % (nchan, av_time))
    make_plot(
        RATar,
        DECTar,
        lotss_result_file,
        extreme_catalogue,
        result,
        targRA,
        targDEC,
        nchan=nchan,
        av_time=av_time,
        outdir=outdir,
    )

    if fit_spec:
        fit_spectrum(delay_cals_file, outdir)

    if vlass:
        from plot_field.vlass_search import search_vlass

        ## Get cutouts of all LBCS sources
        print("Getting cutouts of LBCS sources")
        for i, source in enumerate(result):
            ra, dec = source["RA"], source["DEC"]
            c = SkyCoord(ra, dec, unit=(u.deg, u.deg))
            outfile = os.path.join(outdir, "%s_vlass.fits" % source["Observation"])
            if os.path.exists(outfile):
                continue
            try:
                search_vlass(c, crop=True, crop_scale=256)
                os.system("mv vlass_post**.fits  %s" % outfile)
                convert_vlass_fits(outfile)
            except Exception:
                print(f"VLASS Download failed for {source['Observation']}")
            search_vlass(c, crop=True, crop_scale=256)
            os.system("mv vlass_post**.fits  %s" % outfile)
            convert_vlass_fits(outfile)

    if html:
        app = make_html(
            RATar,
            DECTar,
            lotss_result_file,
            extreme_catalogue,
            result,
            targRA,
            targDEC,
            nchan=nchan,
            av_time=av_time,
            pointing=pointing,
        )
        app.run_server(debug=True, use_reloader=False)

    # return


def parse_args():
    parser = argparse.ArgumentParser(
        prog = "plot-field",
        description="Download and plot LOFAR data regarding a field of interest. This provides the information needed to run the PiLOT pipeline.",
    )
    parser.add_argument(
        "--output_dir",
        dest="outdir",
        type=str,
        help="directory to save results in [default cwd]",
        default=".",
    )
    parser.add_argument(
        "--lotss_radius",
        dest="lotss_radius",
        type=float,
        help="Radius to search LoTSS",
        default=1.5,
    )
    parser.add_argument(
        "--lbcs_radius",
        dest="lbcs_radius",
        type=float,
        help="Radius to search LBCS",
        default=1.5,
    )
    parser.add_argument(
        "--im_radius",
        dest="im_radius",
        type=float,
        help="Radius in which to image",
        default=1.24,
    )
    parser.add_argument(
        "--lotss_catalogue",
        dest="lotss_catalogue",
        type=str,
        help="input file for LoTSS catalogue [will be downloaded if does not exist]",
        default="lotss_catalogue.csv",
    )
    parser.add_argument(
        "--lbcs_catalogue",
        dest="lbcs_catalogue",
        type=str,
        help="input file for LBCS catalogue [will be downloaded if does not exist]",
        default="lbcs_catalogue.csv",
    )
    parser.add_argument(
        "--lotss_result_file",
        dest="lotss_result_file",
        type=str,
        help="output file of sources to image",
        default="image_catalogue.csv",
    )
    parser.add_argument(
        "--delay_cals_file",
        dest="delay_cals_file",
        type=str,
        help="output file of delay calibrators",
        default="delay_calibrators.csv",
    )
    parser.add_argument(
        "--match_tolerance",
        dest="match_tolerance",
        type=float,
        help="radius for matching LBCS to LoTSS [arcsec]",
        default=5.0,
    )
    parser.add_argument(
        "--bright_limit_Jy",
        dest="bright_limit_Jy",
        type=float,
        help="Flux limit for bright sources [Jy]",
        default=5.0,
    )
    parser.add_argument(
        "--image_limit_Jy",
        dest="image_limit_Jy",
        type=float,
        help="Flux limit for which sources to image [Jy]",
        default=0.01,
    )
    parser.add_argument(
        "--continue_no_lotss",
        dest="continue_no_lotss",
        action="store_true",
        help="Continue with the pipeline if no lotss cross-matches can be found?",
        default=False,
    )
    parser.add_argument("--targRA", type=float, dest="targRA", help="Target RA in deg")
    parser.add_argument(
        "--targDEC", type=float, dest="targDEC", help="Target DEC in deg"
    )
    parser.add_argument(
        "--nchan",
        dest="nchan",
        type=int,
        help="Number of frequency channels",
        default=16,
    )
    parser.add_argument(
        "--av_time", dest="av_time", type=float, help="Time averaging", default=1.0
    )

    parser.add_argument("--MS", type=str, dest="MS", help="Measurement Set")
    parser.add_argument(
        "--vlass",
        dest="vlass",
        action="store_true",
        help="Get VLASS cutouts of delay_calibraors.csv sources",
        default=False,
    )
    parser.add_argument(
        "--html",
        dest="html",
        action="store_true",
        help="Create html page of plots",
        default=False,
    )

    parser.add_argument(
        "--fit_spec",
        dest="fit_spec",
        action="store_true",
        help="Perform spectral fitting",
        default=False,
    )

    parser.add_argument(
        "--force",
        dest="force",
        action="store_true",
        help="Force overwrite of existing csvs",
        default=False,
    )

    parser.add_argument("--RA", type=float, dest="RA", help="Ptg RA in deg")
    parser.add_argument("--DEC", type=float, dest="DEC", help="Ptg DEC in deg")
    parser.add_argument("--pointing", type=str, dest="pointing", help="Pointing name")

    args = parser.parse_args()
    return args


def main():

    args = parse_args()

    if args.MS is not None:
        print("Using MS to get RA and DEC")
        ptgRA, ptgDEC = grab_coo_MS(args.MS)
    else:
        ptgRA, ptgDEC = args.RA, args.DEC

    generate_catalogues(
        float(ptgRA),
        float(ptgDEC),
        targRA=args.targRA,
        targDEC=args.targDEC,
        lotss_radius=args.lotss_radius,
        lbcs_radius=args.lbcs_radius,
        im_radius=args.im_radius,
        bright_limit_Jy=args.bright_limit_Jy,
        lotss_catalogue=args.lotss_catalogue,
        lbcs_catalogue=args.lbcs_catalogue,
        lotss_result_file=args.lotss_result_file,
        delay_cals_file=args.delay_cals_file,
        match_tolerance=args.match_tolerance,
        image_limit_Jy=args.image_limit_Jy,
        continue_no_lotss=args.continue_no_lotss,
        nchan=args.nchan,
        av_time=args.av_time,
        vlass=args.vlass,
        html=args.html,
        outdir=args.outdir,
        pointing=args.pointing,
        fit_spec=args.fit_spec,
        force=args.force
    )

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--output_dir",
        dest="outdir",
        type=str,
        help="directory to save results in [default cwd]",
        default=".",
    )
    parser.add_argument(
        "--lotss_radius",
        dest="lotss_radius",
        type=float,
        help="Radius to search LoTSS",
        default=1.5,
    )
    parser.add_argument(
        "--lbcs_radius",
        dest="lbcs_radius",
        type=float,
        help="Radius to search LBCS",
        default=1.5,
    )
    parser.add_argument(
        "--im_radius",
        dest="im_radius",
        type=float,
        help="Radius in which to image",
        default=1.24,
    )
    parser.add_argument(
        "--lotss_catalogue",
        dest="lotss_catalogue",
        type=str,
        help="input file for LoTSS catalogue [will be downloaded if does not exist]",
        default="lotss_catalogue.csv",
    )
    parser.add_argument(
        "--lbcs_catalogue",
        dest="lbcs_catalogue",
        type=str,
        help="input file for LBCS catalogue [will be downloaded if does not exist]",
        default="lbcs_catalogue.csv",
    )
    parser.add_argument(
        "--lotss_result_file",
        dest="lotss_result_file",
        type=str,
        help="output file of sources to image",
        default="image_catalogue.csv",
    )
    parser.add_argument(
        "--delay_cals_file",
        dest="delay_cals_file",
        type=str,
        help="output file of delay calibrators",
        default="delay_calibrators.csv",
    )
    parser.add_argument(
        "--match_tolerance",
        dest="match_tolerance",
        type=float,
        help="radius for matching LBCS to LoTSS [arcsec]",
        default=5.0,
    )
    parser.add_argument(
        "--bright_limit_Jy",
        dest="bright_limit_Jy",
        type=float,
        help="Flux limit for bright sources [Jy]",
        default=5.0,
    )
    parser.add_argument(
        "--image_limit_Jy",
        dest="image_limit_Jy",
        type=float,
        help="Flux limit for which sources to image [Jy]",
        default=0.01,
    )
    parser.add_argument(
        "--continue_no_lotss",
        dest="continue_no_lotss",
        action="store_true",
        help="Continue with the pipeline if no lotss cross-matches can be found?",
        default=False,
    )
    parser.add_argument("--targRA", type=float, dest="targRA", help="Target RA in deg")
    parser.add_argument(
        "--targDEC", type=float, dest="targDEC", help="Target DEC in deg"
    )
    parser.add_argument(
        "--nchan",
        dest="nchan",
        type=int,
        help="Number of frequency channels",
        default=16,
    )
    parser.add_argument(
        "--av_time", dest="av_time", type=float, help="Time averaging", default=1.0
    )

    parser.add_argument("--MS", type=str, dest="MS", help="Measurement Set")
    parser.add_argument(
        "--vlass",
        dest="vlass",
        action="store_true",
        help="Get VLASS cutouts of delay_calibraors.csv sources",
        default=False,
    )
    parser.add_argument(
        "--html",
        dest="html",
        action="store_true",
        help="Create html page of plots",
        default=False,
    )

    parser.add_argument(
        "--fit_spec",
        dest="fit_spec",
        action="store_true",
        help="Perform spectral fitting",
        default=False,
    )

    parser.add_argument(
        "--force",
        dest="force",
        action="store_true",
        help="Force overwrite of existing csvs",
        default=False,
    )

    parser.add_argument("--RA", type=float, dest="RA", help="Ptg RA in deg")
    parser.add_argument("--DEC", type=float, dest="DEC", help="Ptg DEC in deg")
    parser.add_argument("--pointing", type=str, dest="pointing", help="Pointing name")

    args = parser.parse_args()

    if args.MS is not None:
        print("Using MS to get RA and DEC")
        ptgRA, ptgDEC = grab_coo_MS(args.MS)
    else:
        ptgRA, ptgDEC = args.RA, args.DEC

    generate_catalogues(
        float(ptgRA),
        float(ptgDEC),
        targRA=args.targRA,
        targDEC=args.targDEC,
        lotss_radius=args.lotss_radius,
        lbcs_radius=args.lbcs_radius,
        im_radius=args.im_radius,
        bright_limit_Jy=args.bright_limit_Jy,
        lotss_catalogue=args.lotss_catalogue,
        lbcs_catalogue=args.lbcs_catalogue,
        lotss_result_file=args.lotss_result_file,
        delay_cals_file=args.delay_cals_file,
        match_tolerance=args.match_tolerance,
        image_limit_Jy=args.image_limit_Jy,
        continue_no_lotss=args.continue_no_lotss,
        nchan=args.nchan,
        av_time=args.av_time,
        vlass=args.vlass,
        html=args.html,
        outdir=args.outdir,
        pointing=args.pointing,
        fit_spec=args.fit_spec,
        force=args.force
    )


### TO DO LIST

# Add colour for quality of LBCS sources - Use fit or the compactness codes - Maybe not
